/*	tableLookup.c

	Copyright 2008 Fred Cooke

	This file is part of the FreeEMS project.

	FreeEMS software is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	FreeEMS software is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with any FreeEMS software.  If not, see <http://www.gnu.org/licenses/>.

	We ask that if you make any changes to this file you send them upstream to us at admin@diyefi.org

	Thank you for choosing FreeEMS to run your engine! */

#define TABLELOOKUP_C
#include "inc/freeEMS.h"
#include "inc/commsISRs.h"
#include "inc/tableLookup.h"

/* Take a table with two movable axis sets and two axis lengths,
 * loop to find which pairs of axis values and indexs we are between,
 * interpolate two pairs down to two values,
 * interpolate two values down to one value.
 * 
 * Warning : This function relies on the axis values being a sorted list from low to high. If this is not the case behaviour is undefined and could include memory corruption and engine damage.
 * 
 * To reduce the table size from 19x24 to something smaller, simply redcue the RPMLength and LoadLength fields to lower values. Increasing the size of either axis is not currently possible.
 * 
 * Given that the axis lists are in order, a data point outside the table will give the value adjacent to it, and one outside one of the four corners will give the corner value. This is a clean and reasonable behaviour in my opinion.
 * 
 * Reminder : X/RPM is horizontal, Y/Load is vertical */
unsigned short lookupMainTable(mainTable *Table, unsigned short realRPM, unsigned short realLoad){

	/* Find the bounding axis values and indices for RPM */
	unsigned char lowRPMIndex = 0;
	unsigned char highRPMIndex = Table->RPMLength - 1; // if never set in the loop, low value will equal high value and will be on the edge of the map
	unsigned short lowRPMValue = Table->RPM[0]; // if never set in the loop, low value will equal high value and will be on the edge of the map
	unsigned short highRPMValue = Table->RPM[Table->RPMLength -1]; // if never set in the loop, low value will equal high value and will be on the edge of the map
	unsigned char RPMIndex;
	for(RPMIndex=0;RPMIndex<Table->RPMLength;RPMIndex++){
		if(Table->RPM[RPMIndex] < realRPM){
			lowRPMValue = Table->RPM[RPMIndex];
			lowRPMIndex = RPMIndex;
		}else if(Table->RPM[RPMIndex] > realRPM){
			highRPMValue = Table->RPM[RPMIndex];
			highRPMIndex = RPMIndex;
			break;
		}else if(Table->RPM[RPMIndex] == realRPM){
			lowRPMValue = Table->RPM[RPMIndex];
			highRPMValue = Table->RPM[RPMIndex];
			lowRPMIndex = RPMIndex;
			highRPMIndex = RPMIndex;
			break;
		}
	}


	/* Find the bounding cell values and indices for Load */
	unsigned char lowLoadIndex = 0;
	unsigned char highLoadIndex = Table->LoadLength -1; // if never set in the loop, low value will equal high value and will be on the edge of the map
	unsigned short lowLoadValue = Table->Load[0]; // if never set in the loop, low value will equal high value and will be on the edge of the map
	unsigned short highLoadValue = Table->Load[Table->LoadLength -1]; // if never set in the loop, low value will equal high value and will be on the edge of the map
	unsigned char LoadIndex;
	for(LoadIndex=0;LoadIndex<Table->LoadLength;LoadIndex++){
		if(Table->Load[LoadIndex] < realLoad){
			lowLoadValue = Table->Load[LoadIndex];
			lowLoadIndex = LoadIndex;
		}else if(Table->Load[LoadIndex] > realLoad){
			highLoadValue = Table->Load[LoadIndex];
			highLoadIndex = LoadIndex;
			break;
		}else if(Table->Load[LoadIndex] == realLoad){
			lowLoadValue = Table->Load[LoadIndex];
			highLoadValue = Table->Load[LoadIndex];
			lowLoadIndex = LoadIndex;
			highLoadIndex = LoadIndex;
			break;
		}
	}


	/* Obtain the four corners surrounding the spot of interest */
	unsigned short lowRPMLowLoad = Table->Table[(Table->LoadLength * lowRPMIndex) + lowLoadIndex];
	unsigned short lowRPMHighLoad = Table->Table[(Table->LoadLength * lowRPMIndex) + highLoadIndex];
	unsigned short highRPMLowLoad = Table->Table[(Table->LoadLength * highRPMIndex) + lowLoadIndex];
	unsigned short highRPMHighLoad = Table->Table[(Table->LoadLength * highRPMIndex) + highLoadIndex];


	/* Find the two side values to interpolate between by iterpolation */
	unsigned short lowRPMIntLoad = lowRPMLowLoad + (((signed long)((signed long)lowRPMHighLoad - lowRPMLowLoad) * (realLoad - lowLoadValue))/ (highLoadValue - lowLoadValue));
	unsigned short highRPMIntLoad = highRPMLowLoad + (((signed long)((signed long)highRPMHighLoad - highRPMLowLoad) * (realLoad - lowLoadValue))/ (highLoadValue - lowLoadValue));

		
	/* Interpolate the final value from the two side values */
	unsigned short intRPMIntLoad = lowRPMIntLoad + (((signed long)((signed long)highRPMIntLoad - lowRPMIntLoad) * (realRPM - lowRPMValue))/ (highRPMValue - lowRPMValue));


	/* Return the value required */
	return intRPMIntLoad;
}

unsigned short lookupTwoDTableUS(twoDTableUS * Table, unsigned short Value){
	return 0;
}


//unsigned char lookup8Bit3dUC(void){
//	// TODO
//}
//
//unsigned short lookup16Bit2dUS(void){
//	// TODO
//}
//
//unsigned char lookup8Bit2dUC(void){
//	// TODO
//}
//
//signed short lookup16Bit3dSS(??
//signed short lookup16Bit3dSS(void){
//	// TODO
//}
//
//signed char lookup8Bit3D(void){
//	// TODO
//}
